from .load import load


def celias(trange=['2006-06-01', '2006-06-02'],
        datatype='pm_5min',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Charge, Element, and Isotope Analysis System (CELIAS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'pm_5min' for L1 data
                'pm_30s' for L1 data
                'sem_1day' for L2 data
                'sem_15s' for L2 data

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='celias', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return celias_postprocessing(tvars)


def celias_postprocessing(variables):
    """
    Placeholder for CELIAS post-processing 
    """
    return variables


def costep(trange=['2006-06-01', '2006-06-02'],
        datatype='ephin_l3i-1day',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Comprehensive Suprathermal and Energetic Particle Analyzer (COSTEP)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; options: (default: ephin_l3i-1day)
                    ephin_l3i-1day
                    ephin_l3i-1hr
                    ephin_l3i-30min
                    ephin_l3i-10min
                    ephin_l3i-5min
                    ephin_l3i-1min
        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='costep', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return costep_postprocessing(tvars)


def costep_postprocessing(variables):
    """
    Placeholder for COSTEP post-processing 
    """
    return variables


def erne(trange=['2006-06-01', '2006-06-02'],
        datatype='hed_l2-1min',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Energetic and Relativistic Nuclei and Electron experiment (ERNE)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'hed_l2-1min' for L2 data
                'led_l2-1min' for L2 data

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='erne', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return erne_postprocessing(tvars)


def erne_postprocessing(variables):
    """
    Placeholder for ERNE post-processing 
    """
    return variables


def orbit(trange=['2006-06-01', '2006-06-02'],
        datatype='pre_or',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Orbit (ephemeris and attitude) data (ORBIT)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'def_at' for definitive attitude data
                'def_or' for definitive orbit data
                'pre_or' for preliminary orbit data (default)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars= load(instrument='orbit', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return orbit_postprocessing(tvars)


def orbit_postprocessing(variables):
    """
    Placeholder for ORBIT post-processing 
    """
    return variables

        




import os

CONFIG = {'local_data_dir': 'erg_data/',
          'remote_data_dir': 'https://ergsc.isee.nagoya-u.ac.jp/data/ergsc/'}

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join(
        [os.environ['SPEDAS_DATA_DIR'], 'ergsc'])

if os.environ.get('ERG_DATA_DIR'):
    CONFIG['local_data_dir'] = os.environ['ERG_DATA_DIR']

if os.environ.get('ERG_REMOTE_DATA_DIR'):
    CONFIG['remote_data_dir'] = os.environ['ERG_REMOTE_DATA_DIR']



import logging
import numpy as np
from pytplot import get_data
from pyspedas import time_string


def tplot_average(tvar, trange, quiet=False):
    """
    Returns the average value of a tplot variable over a specified time range.

    Input
    -----
        tvar: str
            Name of the tplot variable to average

        trange: list of str or list of float
            Time range to average over

    Returns
    -------
        Average value of the tplot variable
    """
    data = get_data(tvar, dt=True)

    if data is None:
        logging.error('Error reading: ' + tvar)
        return

    if len(trange) != 2:
        logging.error('Error: time range must be two element array.')
        return

    trange = time_string(trange)

    t0 = np.datetime64(trange[0])
    t1 = np.datetime64(trange[-1])

    if not quiet:
        logging.info('Averaging ' + tvar)

    # find the data within the time range
    indices = np.argwhere((data.times <= t1) & (data.times >= t0))

    if len(indices) != 0:
        if len(data.y.shape) > 1:
            return np.array([np.nanmean(data.y[indices, 0]), np.nanmean(data.y[indices, 1]), np.nanmean(data.y[indices, 2])])
        else:
            return np.nanmean(data.y[indices])



import cdflib
import numpy as np

from copy import deepcopy
from pytplot import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load


def isee_brio(
    trange=['2020-08-01', '2020-08-02'],
    suffix='',
    site='all',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True
):

    #;----- datatype -----;
    datatype='64hz'
    instr='brio'
    freq='30'
    
    site_code_all = ['ath', 'kap', 'gak', 'hus', 'zgn', 'ist']
    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = site_code_all
    
    site_code = list(set(site_code).intersection(site_code_all))

    prefix = 'iseetmp_'
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        fres = datatype
        file_res = 3600. * 24
        pathformat = 'ground/riometer/'+site_input\
                        +'/%Y/isee_'+fres+'_'+instr+freq+'_'+site_input+'_%Y%m%d_v??.cdf'
        
        loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, datatype=datatype, prefix=prefix, suffix='_' + site_input +suffix, get_support_data=get_support_data,
                        varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
        
        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                if isinstance(loaded_data_temp, list):
                    if downloadonly:
                        cdf_file = cdflib.CDF(loaded_data_temp[-1])
                        gatt = cdf_file.globalattsget()
                    else:
                        gatt = get_data(loaded_data_temp[-1], metadata=True)['CDF']['GATT']
                elif isinstance(loaded_data_temp, dict):
                    gatt = loaded_data_temp[list(loaded_data_temp.keys())[-1]]['CDF']['GATT']
                print('**************************************************************************')
                print(gatt["Logical_source_description"])
                print('')
                print(f'Information about {gatt["Station_code"]}')
                print(f'PI {gatt["PI_name"]}')
                print('')
                print(f'Affiliations: {gatt["PI_affiliation"]}')
                print('')
                print('Rules of the Road for ISEE Riometer Data:')
                print('')
                print(gatt["TEXT"])
                print({gatt["LINK_TEXT"]})
                print('**************************************************************************')
            except:
                print('printing PI info and rules of the road was failed')
            
        if (not downloadonly) and (not notplot):
            if len(loaded_data_temp) > 0:
                file_name = get_data(loaded_data_temp[-1], metadata=True)['CDF']['FILENAME']
                if isinstance(file_name, list):
                    file_name = file_name[0]
                cdf_file = cdflib.CDF(file_name)
                cdf_info = cdf_file.cdf_info()
                all_cdf_variables = cdf_info['rVariables'] + cdf_info['zVariables']
                for t_plot_name in loaded_data_temp:
                    get_data_vars = get_data(t_plot_name)
                    if get_data_vars is None:
                        store_data(t_plot_name, delete=True)
                    else:
                        t_plot_name_split =t_plot_name.split('_')
                        if len(t_plot_name_split) > 2:
                            #;----- Find param -----;
                            param = t_plot_name_split[1]
                            if param in ['cna', 'qdc', 'raw']:
                                #;----- Rename tplot variables -----;
                                new_tplot_name = 'isee_brio'+freq+'_'+site_input+'_'+fres+'_'+param + suffix
                                store_data(t_plot_name, newname=new_tplot_name)
                                loaded_data.remove(t_plot_name)
                                loaded_data.append(new_tplot_name)
                                #;----- Missing data -1.e+31 --> NaN -----;
                                clip(new_tplot_name, -1e+5, 1e+5)
                                get_data_vars = get_data(new_tplot_name)
                                if param in all_cdf_variables:
                                    var_atts = cdf_file.varattsget(param)
                                    if "FILLVAL" in var_atts:  #removing "FILLVAL", like 999.9000 or 0.0.
                                                            #removing process of cdf_to_tplot.py may be not working well.
                                        var_properties = cdf_file.varinq(param)
                                        if ((var_properties['Data_Type_Description'] ==
                                                'CDF_FLOAT') or
                                                (var_properties['Data_Type_Description'] ==
                                                'CDF_REAL4') or
                                                (var_properties['Data_Type_Description'] ==
                                                'CDF_DOUBLE') or
                                                (var_properties['Data_Type_Description'] ==
                                                'CDF_REAL8')):
                                            if isinstance(var_atts["FILLVAL"], str):
                                                fill_value = float(var_atts["FILLVAL"])
                                            else:
                                                fill_value = deepcopy(var_atts["FILLVAL"])
                                            new_y = np.where(get_data_vars[1] == fill_value,
                                                            np.nan, get_data_vars[1])
                                            get_metadata_vars = get_data(new_tplot_name, metadata=True)
                                            store_data(new_tplot_name, data={'x':get_data_vars[0],
                                                                                'y':new_y},
                                                        attr_dict=get_metadata_vars)
                                            ylim(new_tplot_name, np.nanmin(new_y), np.nanmax(new_y))
                                    else:
                                        ylim(new_tplot_name, np.nanmin(get_data_vars[1]), np.nanmax(get_data_vars[1]))
                                #;----- Set options -----;
                                options(new_tplot_name, 'ytitle', site_input.upper())
                                if param == 'cna':
                                    options(new_tplot_name, 'ysubtitle', '[dB]')
                                    options(new_tplot_name, 'legend_names', ['CNA'])
                                elif param == 'qdc':
                                    options(new_tplot_name, 'ysubtitle', '[V]')
                                    options(new_tplot_name, 'legend_names', ['QDC'])
                                elif param == 'raw':
                                    options(new_tplot_name, 'ysubtitle', '[V]')
                                    options(new_tplot_name, 'legend_names', ['Raw data'])

    return loaded_data



import numpy as np


def erg_pgs_limit_range(data_in, phi=None, theta=None, energy=None,
                        no_ang_weighting=False):

    """
    Applies phi, theta, and energy limits to data structure(s) by
    turning off the corresponding bin flags.

    Input:
        data: dict
            Particle data structure

    Parameters:
        phi: np.ndarray
            Minimum and maximum values for phi

        theta: np.ndarray
            Minimum and maximum values for theta

        energy: np.ndarray
            Minimum and maximum values for energy

    Returns:
        Data structure with limits applied (to the bins array)
    """

    # Apply phi limits

    if phi is not None:

        # phi can be in any order
        phi_in = phi

        # get min/max phi values for all bins
        if not no_ang_weighting:
            phi_min = data_in['phi'] - 0.5 * data_in['dphi']
            phi_max = np.fmod(data_in['phi'] + 0.5 * data_in['dphi'], 360.)
        else:
            phi_min = data_in['phi'] - 0.
            phi_max = np.fmod(data_in['phi'] + 0., 360.)

        # ;wrap negative values
        phi_min[phi_min < 0.0] += 360.

        """
        ;the code below and the phi spectrogram code
        ;assume maximums at 360 are not wrapped to 0
        """
        phi_max[phi_max == 0.0] = 360.0

        # ;find which bins were wrapped back into [0,360]
        wrapped = phi_min > phi_max

        # determine which bins intersect the specified range
        if phi_in[0] > phi_in[1]:
            in_range = phi_min < phi_in[1] or phi_max > phi_in[0] or wrapped
        else:
            in_range = ((phi_min < phi_in[1]) & (phi_max > phi_in[0]))\
                        | (wrapped & ((phi_min < phi_in[1]) | (phi_max > phi_in[0])))

        data_in['bins'][in_range == False] = 0

    # ;Apply theta limits
    if theta is not None:

        input_theta_array = np.array(theta)
        theta_min_max = np.array([np.nanmin(input_theta_array),
                                  np.nanmax(input_theta_array)])
        if not no_ang_weighting:
            theta_min = data_in['theta'] - 0.5 * data_in['dtheta']
            theta_max = data_in['theta'] + 0.5 * data_in['dtheta']
        else:
            theta_min = data_in['theta'] - 0.
            theta_max = data_in['theta'] + 0.

        # ;determine which bins intersect the specified range
        in_range = ((theta_min < theta_min_max[1]) \
                    & (theta_max > theta_min_max[0]))
        data_in['bins'][in_range == False] = 0

    # ;Apply energy limits
    if energy is not None:
        data_in['bins'][data_in['energy'] < energy[0]] = 0
        data_in['bins'][data_in['energy'] > energy[1]] = 0

    return data_in



import logging
import pandas as pd
import numpy as np
from pytplot import time_clip as tclip
from pytplot import store_data, options


def mms_load_eph_tplot(filenames, level='def', probe='1', datatypes=['pos', 'vel'], suffix='', trange=None):
    """
    Helper routine for loading state data (ASCII files from the SDC); not meant to be called directly; see pyspedas.mms.state instead
    """
    prefix = 'mms' + probe

    time_values = []
    date_values = []
    x_values = []
    y_values = []
    z_values = []
    vx_values = []
    vy_values = []
    vz_values = []

    for file in filenames:
        logging.info('Loading ' + file)
        rows = pd.read_csv(file, delim_whitespace=True, header=None, skiprows=14)
        times = rows.shape[0]-1
        for time_idx in range(0, times):
            # these files can overlap, so avoid duplicates
            if rows[0][time_idx] in date_values:
                continue
            time_values.append(pd.to_datetime(rows[0][time_idx], format='%Y-%j/%H:%M:%S.%f').timestamp())
            x_values.append(rows[2][time_idx])
            y_values.append(rows[3][time_idx])
            z_values.append(rows[4][time_idx])
            vx_values.append(rows[5][time_idx])
            vy_values.append(rows[6][time_idx])
            vz_values.append(rows[7][time_idx])
            date_values.append(rows[0][time_idx])

    if 'pos' in datatypes:
        store_data(prefix + '_' + level + 'eph_pos' + suffix, data={'x': time_values, 'y': np.transpose(np.array([x_values, y_values, z_values]))})
        tclip(prefix + '_' + level + 'eph_pos' + suffix, trange[0], trange[1], suffix='')
        options(prefix + '_' + level + 'eph_pos' + suffix, 'ytitle', 'MMS'+str(probe)+' position')
        options(prefix + '_' + level + 'eph_pos' + suffix, 'ysubtitle', '[km]')
        options(prefix + '_' + level + 'eph_pos' + suffix, 'legend_names', ['X ECI', 'Y ECI', 'Z ECI'])

    if 'vel' in datatypes:
        store_data(prefix + '_' + level + 'eph_vel' + suffix, data={'x': time_values, 'y': np.transpose(np.array([vx_values, vy_values, vz_values]))})
        tclip(prefix + '_' + level + 'eph_vel' + suffix, trange[0], trange[1], suffix='')
        options(prefix + '_' + level + 'eph_vel' + suffix, 'ytitle', 'MMS'+str(probe)+' velocity')
        options(prefix + '_' + level + 'eph_vel' + suffix, 'ysubtitle', '[km/s]')
        options(prefix + '_' + level + 'eph_vel' + suffix, 'legend_names', ['Vx ECI', 'Vy ECI', 'Vz ECI'])



import numpy as np

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum


def mms_pgs_make_theta_spec(data_in, resolution=16, colatitude=False):
    """
    Builds a theta (latitudinal) spectrogram from a simplified particle data structure.

    Parameters
    ----------
    data_in : dict
        A dictionary containing the particle data, including 'data', 'theta', and 'bins' keys.
    resolution : int, optional
        The number of bins to use for the spectrogram. Defaults to 16.
    colatitude : bool, optional
        Set to True if the input data is in colatitude rather than latitude.

    Returns
    -------
    y : numpy.ndarray
        The y axis of the spectrogram.
    ave : numpy.ndarray
        The spectrogram.
    """
    data = data_in.copy()
    n_theta = resolution

    # zero inactive bins to ensure areas with no data are represented as NaN
    zero_bins = np.argwhere(data['bins'] == 0)
    if zero_bins.size != 0:
        for item in zero_bins:
            data['data'][item[0], item[1]] = 0.0

    ave = np.zeros(n_theta)
    bin_size = 180.0/n_theta
    outbins = np.arange(0, 181.0, bin_size)

    # shift to colatitude
    if not colatitude:
        data['theta'] = 90.0-data['theta']

    theta_flat = data['theta'].flatten()
    data_flat = data['data'].flatten()
    bins_flat = data['bins'].flatten()

    for bin_idx in range(0, len(outbins)-1):
        this_bin = np.argwhere((theta_flat >= outbins[bin_idx]) & (theta_flat < outbins[bin_idx+1]))
        if len(this_bin) > 0:
            bins = nansum(bins_flat[this_bin])
            if bins != 0.0:
                ave[bin_idx] += nansum(data_flat[this_bin])/bins

    if not colatitude:
        data['theta'] = 90.0-data['theta']
        outbins = 90.0-outbins

    y = outbins[0:n_theta]+0.5*(outbins[1::]-outbins[0:n_theta])

    return y, ave



import numpy as np

from pyspedas import tnames, tcopy
from pytplot import get_data, store_data

from pyspedas.cotrans.cotrans import cotrans
from pytplot import tnormalize
from pytplot import tcrossp
from pyspedas.analysis.tinterpol import tinterpol

from ..common.cotrans.erg_cotrans import erg_cotrans

# ;so we don't have one long routine of doom, all transforms should be separate helper functions
def erg_pgs_xgse(
    mag_temp,
    pos_tmp=None
):
    
    mag_data = get_data(mag_temp)

    # xaxis of this system is X of the gse system. Z is mag field
    x_axis = np.zeros((len(mag_data.times), 3))
    x_axis[:, 0] = 1
    
    store_data('xgse_pgs_temp', data={'x':mag_data[0], 'y':x_axis})
    cotrans('xgse_pgs_temp', name_out='xgse_pgs_temp',
            coord_in='gse', coord_out='j2000')
    erg_cotrans(in_name='xgse_pgs_temp', out_name='xgse_pgs_temp',
                in_coord='j2000', out_coord='dsi')
    
    # ;create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    y_basis = tcrossp(z_basis, 'xgse_pgs_temp', return_data=True)
    y_basis = tnormalize(y_basis, return_data=True)
    x_basis = tcrossp(y_basis, z_basis, return_data=True)
    
    return (x_basis, y_basis, z_basis)

# ;so we don't have one long routine of doom, all transforms should be separate helper functions
def erg_pgs_phigeo(
    mag_temp,
    pos_temp
):
    
    postmp = 'pos_pgs_temp'
    tcopy(pos_temp, postmp)
    cotrans(postmp, postmp, coord_in='gse', coord_out='geo')
    pos_data = get_data(postmp)
    
    # transformation to generate other_dim dim for phigeo from thm_fac_matrix_make
    # All the conversions to polar and trig simplifies to this.
    # But the reason the conversion is why this is the conversion that is done, is lost on me.
    # The conversion swaps the x & y components of position, reflects over x=0,z=0 then projects into the xy plane
    pos_conv = np.stack((-pos_data.y[:, 1], pos_data.y[:, 0], np.zeros(len(pos_data.times))))
    pos_conv = np.transpose(pos_conv, [1, 0])
    store_data(postmp, data={'x': pos_data.times, 'y': pos_conv})

    # ;transform into dsl because particles are in dmpa
    cotrans(postmp,postmp,coord_in='geo', coord_out='j2000')
    erg_cotrans(postmp,postmp,in_coord='j2000', out_coord='dsi')
    
    # ;create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    x_basis = tcrossp(postmp,z_basis, return_data=True)
    x_basis = tnormalize(x_basis, return_data=True)
    y_basis = tcrossp(z_basis, x_basis, return_data=True)

    return (x_basis, y_basis, z_basis)

# ;so we don't have one long routine of doom, all transforms should be separate helper functions
def erg_pgs_mphigeo(
    mag_temp,
    pos_temp
):
    
    postmp = 'pos_pgs_temp'
    tcopy(pos_temp, postmp)
    cotrans(postmp, postmp, coord_in='gse', coord_out='geo')
    pos_data = get_data(postmp)

    # the following is heisted from the IDL version
    # transformation to generate other_dim dim for mphigeo from thm_fac_matrix_make
    # All the conversions to polar and trig simplifies to this.  
    # But the reason the conversion is why this is the conversion that is done, is lost on me.
    # The conversion swaps the x & y components of position, reflects over x=0,z=0 then projects into the xy plane
    pos_conv = np.stack((pos_data.y[:, 1], -pos_data.y[:, 0], np.zeros(len(pos_data.times))))
    pos_conv = np.transpose(pos_conv, [1, 0])
    store_data(pos_temp, data={'x': pos_data.times, 'y': pos_conv})
    
    # ;transform into dsl because particles are in dmpa
    cotrans(postmp,postmp,coord_in='geo', coord_out='j2000')
    erg_cotrans(postmp,postmp,in_coord='j2000', out_coord='dsi')
    
    # ;create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    x_basis = tcrossp(z_basis, postmp, return_data=True)
    x_basis = tnormalize(x_basis, return_data=True)
    y_basis = tcrossp(z_basis, x_basis, return_data=True)
    
    return (x_basis, y_basis, z_basis)


"""
;; ERG_PGS_PHISM
;; To get unit vectors for FAC-phi_sm coordinate system
;; Z axis: local B-vector dir, usually taken from spin-averaged MGF data
;; X axis: phi_sm vector x Z, where the phi_sm lies in the azimuthally
;;         eastward direction at a spacecraft position in the SM coordinate
;;         system. In a dipole B-field geometry, X axis roughly points
;;         radially outward. 
;; Y axis: Z axis x X axis, roughly pointing azimuthally eastward
;;
;; Common to the other procedures here, pos_temp is the name of a
;; tplot variable containing spacecraft's position coordinates
;; in GSE. "mag_temp" is for the local magnetic field vectors in DSI. 
"""

def erg_pgs_phism(
    mag_temp,
    pos_temp
):
    postmp = 'pos_pgs_temp'
    tcopy(pos_temp, postmp)
    cotrans(postmp, postmp, coord_in='gse', coord_out='sm')
    pos_data = get_data(postmp)
    
    """
    ;; The conversion swaps the x & y components of position, reflects
    ;; over x=0,z=0 then projects into the xy plane. In other words, the
    ;; position vectors projected on the SM-X-Y plane are rotated by +90
    ;; degrees around the SM-Z axis to get the phi_sm vectors.
    """
    
    phitmp = 'phism_tmp'
    pos_conv = np.stack((-pos_data.y[:, 1], pos_data.y[:, 0], np.zeros(len(pos_data.times))))
    pos_conv = np.transpose(pos_conv, [1, 0])
    store_data(phitmp, data={'x': pos_data.times, 'y': pos_conv})
    #   SM to DSI 
    cotrans(phitmp, phitmp, coord_in='sm', coord_out='j2000')
    erg_cotrans(phitmp, phitmp, in_coord='j2000', out_coord='dsi')
    
    # ;; create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    x_basis = tcrossp(phitmp, z_basis, return_data=True)
    x_basis = tnormalize(x_basis, return_data=True)
    y_basis = tcrossp(z_basis, x_basis, return_data=True)
    
    #  ;; clean up the temporary variables
    store_data( [ postmp, phitmp ], delete=True)

    return (x_basis, y_basis, z_basis)

"""
;; ERG_PGS_MPHISM
;; To get unit vectors for FAC-(minus phi_sm) coordinate system
;; Z axis: local B-vector dir, usually taken from spin-averaged MGF data
;; X axis: minus phi_sm vector x Z, where the minus phi_sm lies in the azimuthally
;;         westward direction at a spacecraft position in the SM coordinate
;;         system. In a dipole B-field geometry, X axis roughly points
;;         radially inward. 
;; Y axis: Z axis x X axis, roughly pointing azimuthally westward
;;
;; Common to the other procedures here, pos_temp is the name of a
;; tplot variable containing spacecraft's position coordinates
;; in GSE. "mag_temp" is for the local magnetic field vectors in DSI. 
"""
def erg_pgs_mphism(
    mag_temp,
    pos_temp
):
    
    postmp = 'pos_pgs_temp'
    tcopy(pos_temp, postmp)
    cotrans(postmp, postmp, coord_in='gse', coord_out='sm')
    pos_data = get_data(postmp)
    
    """
    ;; The conversion swaps the x & y components of position, reflects
    ;; over x=0,z=0 then projects into the xy plane. In other words, the
    ;; position vectors projected on the SM-X-Y plane are rotated by +90
    ;; degrees around the SM-Z axis to get the phi_sm vectors.
    """
    phitmp = 'phism_tmp'  # ;; Note that the same name is used as in erg_pgs_phism(), but it has opposite vectors 
    pos_conv = np.stack((pos_data.y[:, 1], -pos_data.y[:, 0], np.zeros(len(pos_data.times))))
    pos_conv = np.transpose(pos_conv, [1, 0])
    store_data(phitmp, data={'x': pos_data.times, 'y': pos_conv})
    #  ;; SM to DSI
    cotrans(phitmp, phitmp, coord_in='sm', coord_out='j2000')
    erg_cotrans(phitmp, phitmp, in_coord='j2000', out_coord='dsi')
    
    # ;; create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    x_basis = tcrossp(phitmp, z_basis, return_data=True)
    x_basis = tnormalize(x_basis, return_data=True)
    y_basis = tcrossp(z_basis, x_basis, return_data=True)

    #  ;; clean up the temporary variables
    store_data( [ postmp, phitmp ], delete=True)

    return (x_basis, y_basis, z_basis)

def erg_pgs_xdsi(
    mag_temp,
    pos_tmp=None
):
    
    mag_data = get_data(mag_temp)

    # xaxis of this system is X of the gse system. Z is mag field
    x_axis = np.zeros((len(mag_data.times), 3))
    x_axis[:, 0] = 1.
    
    # ;create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    y_basis = tcrossp(z_basis, x_axis, return_data=True)
    y_basis = tnormalize(y_basis, return_data=True)
    x_basis = tcrossp(y_basis, z_basis, return_data=True)
    
    return (x_basis, y_basis, z_basis)

def erg_pgs_make_fac(
    times_array,
    mag_tvar_in=None,
    pos_tvar_in=None,
    fac_type='mphism'
):
    """
    Args:
        times_array (Numpy Array): ;the time grid of the particle data.
        mag_tvar_in (str): ;tplot variable containing the mag data in DSI. Defaults to None.
        pos_tvar_in (str, optional): ;position variable containing the position data in GSE. Defaults to None.
        fac_type (str, optional): ;field aligned coordinate transform type (only mphigeo, atm). Defaults to 'mphism'.
    """
    
    valid_types = ['mphigeo', 'phigeo', 'xgse', 'phism', 'mphism', 'xdsi']
    
    if fac_type not in valid_types:
        print(f'Invalid FAC type "{fac_type}"; valid types: {valid_types}')
        return
    
    if (mag_tvar_in is None) or (pos_tvar_in is None):
        print('Magnetic field and/or spacecraft position data not specified.')
        print('Please use mag_tvar_in and pos_tvar_in arguments.')
        return
    
    """
    ;;--------------------------------------------------------------------       
    ;;sanitize
    ;;--------------------------------------------------------------------
    
    ;;Note this logic could probably be rolled into
    ;;thm_pgs_clean_support in the future
    ;; Normnally mag_tvar_in should be erg_mgf_l2_mag_8sec_dsi
    """
    
    if len(tnames(mag_tvar_in)) > 0:
        mag_tvar = tnames(mag_tvar_in)[0]
        mag_temp = mag_tvar + '_pgs_temp'
        #  ;;Right now, magnetic field must be in DSI coordinates
        tcopy(mag_tvar, mag_temp)  # ;;Sanitize it
        tinterpol(mag_temp, times_array, newname=mag_temp)
    else:
        print(f'Magnetic field variable not found: "{mag_tvar_in}"')
        print('skipping field-aligned outputs')
        return
    
    #  ;; Normally pos_tvar_in should be erg_orb_l2_pos_gse
    if len(tnames(pos_tvar_in)) > 0:
        pos_tvar = tnames(pos_tvar_in)[0]
        pos_temp = pos_tvar + '_pgs_temp'
        tcopy(pos_tvar, pos_temp)  # ;;Sanitize it
        tinterpol(pos_temp, times_array, newname=pos_temp)
    else:
        print(f'Position variable not found: "{pos_tvar_in}"')
        print('skipping field-aligned outputs')
        return
    
    if fac_type == 'mphigeo':
        
        """
        ;;--------------------------------------------------------------------
        ;;mphigeo
        ;;--------------------------------------------------------------------
        """
    
        basis = erg_pgs_mphigeo(mag_temp, pos_temp)
        
    elif fac_type == 'phigeo':
        
        """
        ;;--------------------------------------------------------------------
        ;;phigeo
        ;;--------------------------------------------------------------------
        """
        
        basis = erg_pgs_phigeo(mag_temp, pos_temp)
        
    elif fac_type == 'xgse':
        
        """
        ;;--------------------------------------------------------------------
        ;;xgse
        ;;--------------------------------------------------------------------
        """
        
        #  ;;position isn't necessary for this one, but uniformity of interface and requirements trumps here
        basis = erg_pgs_xgse(mag_temp, pos_temp)
        
    elif fac_type == 'phism':
        
        """
        ;;--------------------------------------------------------------------
        ;;phism
        ;;--------------------------------------------------------------------
        """
        
        basis = erg_pgs_phism(mag_temp, pos_temp)
        
    elif fac_type == 'mphism':
        
        """
        ;;--------------------------------------------------------------------
        ;;mphism
        ;;--------------------------------------------------------------------
        """
        
        basis = erg_pgs_mphism(mag_temp,pos_temp)
        
    elif fac_type == 'xdsi':
        
        """
        ;;--------------------------------------------------------------------
        ;;xdsi
        ;;--------------------------------------------------------------------
        """
        
        basis = erg_pgs_xdsi(mag_temp,pos_temp)
        
    """
    ;;--------------------------------------------------------------------
    ;;create rotation matrix
    ;;--------------------------------------------------------------------
    """
    
    fac_output = np.zeros((len(times_array), 3, 3))
    fac_output[:, 0, :] = basis[0]
    fac_output[:, 1, :] = basis[1]
    fac_output[:, 2, :] = basis[2]

    return fac_output